import argparse
import os
import pickle
from numbers import Number
from typing import Optional

import numpy as np
import torch
from torch.optim.optimizer import Optimizer
from tqdm import tqdm

import boilr.data
from boilr.nn.utils import print_num_params
from boilr.options import get_option
from boilr.utils.meta import ObjectWithArgparsedArgs
from boilr.utils.summarize import SummarizerCollection
from boilr.utils.viz import save_image_grid, save_image_grid_reconstructions


class BaseExperimentManager(ObjectWithArgparsedArgs):
    """Base class for experiment manager.

    If `args` is not given, all config parameters are set based on experiment
    defaults and user input, using `argparse`.

    Args:
        args (argparse.Namespace, optional): Experiment configuration.
    """

    def __init__(self, args: Optional[argparse.Namespace] = None):
        super(BaseExperimentManager, self).__init__(args=args)

        self._dataloaders = None
        self._model = None
        self._optimizer = None
        self._run_description = self._make_run_description(self.args)

    @classmethod
    def _define_args_defaults(cls) -> dict:
        defaults = super(BaseExperimentManager, cls)._define_args_defaults()
        defaults.update(
            batch_size=None,
            test_batch_size=None,
            lr=None,
            max_grad_norm=None,
            max_epochs=10000000,
            max_steps=10000000000,
            seed=54321,
            train_log_every=1000,
            test_log_every=1000,
            checkpoint_every=10000,
            keep_checkpoint_max=3,
            resume="",
        )
        return defaults

    def _add_args(self, parser: argparse.ArgumentParser) -> None:

        super(BaseExperimentManager, self)._add_args(parser)

        parser.add_argument('--batch-size',
                            type=int,
                            default=self._default_args['batch_size'],
                            metavar='N',
                            dest='batch_size',
                            help='training batch size')

        parser.add_argument('--test-batch-size',
                            type=int,
                            default=self._default_args['test_batch_size'],
                            metavar='N',
                            dest='test_batch_size',
                            help='test batch size')

        parser.add_argument('--lr',
                            type=float,
                            default=self._default_args['lr'],
                            metavar='LR',
                            dest='lr',
                            help='learning rate')

        parser.add_argument('--max-grad-norm',
                            type=float,
                            default=self._default_args['max_grad_norm'],
                            metavar='NORM',
                            dest='max_grad_norm',
                            help='maximum global norm of the gradient '
                            '(clipped if larger)')

        parser.add_argument('--seed',
                            type=int,
                            default=self._default_args['seed'],
                            metavar='N',
                            dest='seed',
                            help='random seed')

        parser.add_argument('--tr-log-every',
                            type=int,
                            default=self._default_args['train_log_every'],
                            metavar='N',
                            dest='train_log_every',
                            help='log training metrics every this number of '
                            'training steps')

        parser.add_argument('--ts-log-every',
                            type=int,
                            default=self._default_args['test_log_every'],
                            metavar='N',
                            dest='test_log_every',
                            help="log test metrics every this number of "
                            "training steps. It must be a multiple of "
                            "'--tr-log-every'")

        parser.add_argument('--ts-img-every',
                            type=int,
                            metavar='N',
                            dest='test_imgs_every',
                            help="save test images every this number of "
                            "training steps. It must be a multiple of "
                            "'--ts-log-every'. Default: same as "
                            "'--ts-log-every'")

        parser.add_argument('--checkpoint-every',
                            type=int,
                            default=self._default_args['checkpoint_every'],
                            metavar='N',
                            dest='checkpoint_every',
                            help='save model checkpoint every this number of '
                            'training steps')

        parser.add_argument('--keep-checkpoint-max',
                            type=int,
                            default=self._default_args['keep_checkpoint_max'],
                            metavar='N',
                            dest='keep_checkpoint_max',
                            help='keep at most this number of most recent '
                            'model checkpoints')

        parser.add_argument('--max-steps',
                            type=int,
                            default=self._default_args['max_steps'],
                            metavar='N',
                            dest='max_steps',
                            help='max number of training steps')

        parser.add_argument('--max-epochs',
                            type=int,
                            default=self._default_args['max_epochs'],
                            metavar='N',
                            dest='max_epochs',
                            help='max number of training epochs')

        parser.add_argument('--nocuda',
                            action='store_true',
                            dest='no_cuda',
                            help='do not use cuda')

        parser.add_argument('--descr',
                            type=str,
                            default='',
                            metavar='STR',
                            dest='additional_descr',
                            help='additional description for experiment name')

        parser.add_argument('--dry-run',
                            action='store_true',
                            dest='dry_run',
                            help='do not save anything to disk')

        parser.add_argument('--resume',
                            type=str,
                            metavar='NAME',
                            default=self._default_args['resume'],
                            dest='resume',
                            help="load the run with this name and "
                            "resume training")

    @classmethod
    def _check_args(cls, args: argparse.Namespace) -> argparse.Namespace:
        args = super(BaseExperimentManager, cls)._check_args(args)

        # Default: save images at each test
        if args.test_imgs_every is None:
            args.test_imgs_every = args.test_log_every

        if args.test_imgs_every % args.test_log_every != 0:
            msg = ("'test_imgs_every' must be a multiple of 'test_log_every',"
                   " but current values are {img} and {log}")
            msg = msg.format(img=args.test_imgs_every, log=args.test_log_every)
            raise ValueError(msg)

        # This is not necessary but there's no reason not to have it like this,
        # and it usually looks horrible otherwise.
        if args.test_log_every % args.train_log_every != 0:
            msg = ("'test_log_every' must be a multiple of 'train_log_every',"
                   " but current values are {tr} and {ts}")
            msg = msg.format(tr=args.test_log_every, ts=args.train_log_every)
            raise ValueError(msg)

        if args.max_grad_norm is not None:
            assert args.max_grad_norm > 0.0

        return args

    @staticmethod
    def _make_run_description(args: argparse.Namespace) -> str:
        """Creates a string description of the run.

        It is used in the names of the logging folders. By default it returns
        the empty string.

        Args:
            args (argparse.Namespace): Experiment configuration.

        Returns:
            (str): The run description.
        """
        return ""

    def _make_datamanager(self) -> boilr.data.BaseDatasetManager:
        """Creates a dataset manager to wrap data loaders.

        The dataset manager should subclass `boilr.data.BaseDatasetManager`.

        Returns:
            dataset_manager (boilr.data.BaseDatasetManager)
        """
        raise NotImplementedError

    def _make_model(self) -> torch.nn.Module:
        """Creates a model."""
        raise NotImplementedError

    def _make_optimizer(self) -> Optimizer:
        """Creates the optimizer."""
        raise NotImplementedError

    def setup(self,
              device: torch.device,
              checkpoint_folder: Optional[str] = None) -> None:
        """Sets up the experiment.

        Loads the dataset, creates the model, loads the model weights if a
        checkpoint folder is provided, and creates the optimizer.

        Args:
            device (torch.device)
            checkpoint_folder (str, optional)
        """

        self._device = device

        # If checkpoint folder is given, load model from there to resume
        resume = checkpoint_folder is not None

        # Dataset
        print("Getting dataset ready...")
        self._dataloaders = self._make_datamanager()
        print("Data shape: {}".format(self.dataloaders.data_shape))
        print("Train/test set size: {}/{}".format(
            len(self.dataloaders.train.dataset),
            len(self.dataloaders.test.dataset),
        ))

        # Model
        print("Creating model...")
        self._model = self._make_model().to(self.device)

        # Print model overview
        print_num_params(self.model)

        # Load weights if resuming training
        if resume:
            self.load_model(checkpoint_folder, step=None)

        # Optimizer
        self._optimizer = self._make_optimizer()

    def load_args_from_pickle(self, config_pickle_path: str) -> None:
        """Loads experiment config from file when resuming training.

        Args:
            config_pickle_path (str)
        """
        with open(config_pickle_path, 'rb') as file:
            args = pickle.load(file)
        assert not args.dry_run  # this would not make sense
        self._args = args

    def load_model(self,
                   checkpoint_folder: str,
                   step: Optional[int] = None) -> None:
        """Loads model weights from a checkpoint in the specified folder.

        If step is given, it attempts to load the checkpoint at that step.
        The weights are loaded with `map_location=device`, where device is the
        current device of this experiment.

        Args:
            checkpoint_folder (str)
            step (int, optional)
        """
        self.model.load(checkpoint_folder, self.device, step=step)

    def forward_pass(self,
                     x: torch.Tensor,
                     y: Optional[torch.Tensor] = None) -> dict:
        """Simple single-pass model evaluation.

        It consists of a forward pass and computation of all necessary losses
        and metrics.

        Args:
            x (torch.Tensor): data
            y (torch.Tensor, optional): labels

        Returns:
            metrics (dict)
        """
        raise NotImplementedError

    def post_backward_callback(self) -> None:
        """Callback method, called after backward pass and before step."""
        pass

    @classmethod
    def get_metrics_dict(cls, results: dict) -> dict:
        """Returns metrics given a dictionary of results.

        Given a dict of results, returns a dict of metrics to be given to
        summarizers. Keys are also used as names for tensorboard logging.

        In the base implementation, keys are simply copied and used as names
        for tensorboard.

        Only scalars accepted, non-scalars are discarded. Actually, anything
        that either is a scalar or has the method `item()`. That should include
        Python scalars, numpy scalars, torch scalars, numpy and torch
        arrays/tensors with one element.

        Override to customize translation from results to dictionary of
        scalar metrics.

        Args:
            results (dict)

        Returns:
            metrics_dict (dict)
        """
        metrics_dict = {}
        for k in results:
            x = results[k]
            if isinstance(x, Number):
                metrics_dict[k] = x
                continue
            try:
                metrics_dict[k] = x.item()
            except (AttributeError, ValueError):  # is this enough?
                pass
        return metrics_dict

    @classmethod
    def train_log_str(cls,
                      summaries: dict,
                      step: int,
                      epoch: Optional[int] = None) -> str:
        """Returns log string for training metrics."""
        if get_option('show_progress_bar'):
            s = "       "
        else:
            s = "train: "
        s += "[step {}]".format(step)
        for k in summaries:
            s += "  {key}={value:.5g}".format(key=k, value=summaries[k])
        return s

    @classmethod
    def test_log_str(cls,
                     summaries: dict,
                     step: int,
                     epoch: Optional[int] = None) -> str:
        """Returns log string for test metrics."""
        if get_option('show_progress_bar'):
            s = "       "
        else:
            s = "test : "
        if epoch is not None:
            s += "[step {}, epoch {}]".format(step, epoch)
        for k in summaries:
            s += "  {key}={value:.5g}".format(key=k, value=summaries[k])
        return s

    def test_procedure(self, **kwargs) -> dict:
        """Executes the experiment's test procedure and returns results.

        This typically includes collecting metrics on the test set using
        forward_pass(). For example in variational inference we might be
        interested in repeating this many times to derive the importance-
        weighted ELBO.

        The returned summaries are:
        - passed to test_log_str() to get the string that will be printed to
          output (by default all summaries are printed)
        - saved to the history of test metrics, which is saved to disk
        - logged to tensorboard (with 'validation_' prepended to the key)

        Returns:
            summaries (dict)
        """
        raise NotImplementedError

    def save_images(self, img_folder: str) -> None:
        """Saves test images.

        For example, in VAEs, input and reconstruction pairs, or sample from
        the model prior. Images are meant to be saved to the image folder
        that is automatically created by the Trainer.

        By default, do nothing.

        Args:
            img_folder (str): Folder to store images
        """
        pass

    @property
    def run_description(self) -> str:
        """String description of the run, used e.g. for the log folder."""
        return self._run_description

    @property
    def dataloaders(self) -> boilr.data.BaseDatasetManager:
        """Wrapper for train and test data loaders."""
        return self._dataloaders

    @property
    def model(self) -> torch.nn.Module:
        """Model."""
        return self._model

    @property
    def optimizer(self) -> Optimizer:
        """Optimizer."""
        return self._optimizer

    @property
    def device(self) -> torch.device:
        return self._device


class VAEExperimentManager(BaseExperimentManager):
    """Subclass of experiment manager for variational autoencoders.

    This version:
    - implements a default test_procedure for VAEs,
    - defines command line arguments and specified defaults (which can be
      overridden by subclasses),
    - implements default methods for saving model samples (assuming the model
      is a generative model) and for saving pairs of input-reconstruction
      images (assuming it is an autoencoder).

    See the superclass docs for more details.
    """

    def test_procedure(self, iw_samples: Optional[int] = None):
        """Executes the experiment's test procedure and returns results.

        Collects variational inference metrics on the test set using
        forward_pass(), and repeat to derive the importance-weighted ELBO.

        Args:
            iw_samples (int, optional): number of samples for the importance-
                weighted ELBO. The other metrics are also averaged over all
                these samples, yielding a more accurate estimate.

        Returns:
            summaries (dict)
        """

        # Shorthand
        test_loader = self.dataloaders.test
        step = self.model.global_step
        args = self.args
        n_test = len(test_loader.dataset)

        # If it's time to estimate log likelihood, use many samples.
        # If given, use the given number.
        if iw_samples is None:
            iw_samples = 1
            if step % args.loglikelihood_every == 0 and step > 0:
                iw_samples = args.loglikelihood_samples

        # Setup
        summarizers = SummarizerCollection(mode='sum')
        show_progress = get_option('show_progress_bar')
        if show_progress:
            progress = tqdm(total=len(test_loader) * iw_samples, desc='test ')
        all_elbo_sep = torch.zeros(n_test, iw_samples)

        # Do test
        for batch_idx, (x, y) in enumerate(test_loader):
            for i in range(iw_samples):
                outputs = self.forward_pass(x, y)

                # elbo_sep shape (batch size,)
                i_start = batch_idx * args.test_batch_size
                i_end = (batch_idx + 1) * args.test_batch_size
                all_elbo_sep[i_start:i_end, i] = outputs['elbo_sep'].detach()

                metrics_dict = self.get_metrics_dict(outputs)
                multiplier = (x.size(0) / n_test) / iw_samples
                for k in metrics_dict:
                    metrics_dict[k] *= multiplier
                summarizers.add(metrics_dict)

                if show_progress:
                    progress.update()

        if show_progress:
            progress.close()

        if iw_samples > 1:
            # Shape (test set size,)
            elbo_iw = torch.logsumexp(all_elbo_sep, dim=1)
            elbo_iw = elbo_iw - np.log(iw_samples)

            # Mean over test set (scalar)
            elbo_iw = elbo_iw.mean().item()
            key = 'elbo/elbo_IW_{}'.format(iw_samples)
            summarizers.add({key: elbo_iw})

        summaries = summarizers.get_all(reset=True)
        return summaries

    @classmethod
    def _define_args_defaults(cls) -> dict:
        defaults = super(VAEExperimentManager, cls)._define_args_defaults()
        defaults.update(
            loglikelihood_every=50000,
            loglikelihood_samples=100,
        )
        return defaults

    def _add_args(self, parser: argparse.ArgumentParser) -> None:

        super(VAEExperimentManager, self)._add_args(parser)

        parser.add_argument('--ll-every',
                            type=int,
                            default=self._default_args['loglikelihood_every'],
                            metavar='N',
                            dest='loglikelihood_every',
                            help='evaluate log likelihood every this number '
                            'of training steps')

        parser.add_argument('--ll-samples',
                            type=int,
                            default=self._default_args['loglikelihood_samples'],
                            metavar='N',
                            dest='loglikelihood_samples',
                            help='number of importance-weighted samples to '
                            'evaluate log likelihood')

    @classmethod
    def _check_args(cls, args: argparse.Namespace) -> argparse.Namespace:

        args = super(VAEExperimentManager, cls)._check_args(args)

        if args.loglikelihood_every % args.test_log_every != 0:
            msg = ("'loglikelihood_every' must be a multiple of "
                   "'test_log_every', but current values are {ll} and {log}")
            msg = msg.format(ll=args.loglikelihood_every,
                             log=args.test_log_every)
            raise ValueError(msg)

        return args

    def generate_and_save_samples(self,
                                  filename: str,
                                  nrows: Optional[int] = 8) -> None:
        """Default method to generate and save model samples.

        Args:
            filename (str): file name.
            nrows (int, optional): number of rows (and columns) of images.
        """
        samples = self.model.sample_prior(nrows**2)
        save_image_grid(samples, filename, nrows=nrows)

    def generate_and_save_reconstructions(self,
                                          x: torch.Tensor,
                                          filename: str,
                                          nrows: Optional[int] = 8) -> None:
        """Default method to generate and save input/reconstruction pairs.

        Args:
            x (torch.Tensor): a Tensor of input images (N, channels, H, W)
            filename (str): file name.
            nrows (int, optional): number of rows (and columns) of images.
        """
        n_img = nrows**2 // 2
        if x.shape[0] < n_img:
            msg = ("{} data points required, but given batch has size {}. "
                   "Please use a larger batch.".format(n_img, x.shape[0]))
            raise RuntimeError(msg)
        x = x.to(self.device)
        outputs = self.forward_pass(x)

        # Try to get reconstruction from different sources in order
        recons = None
        possible_recons_names = ['out_recons', 'out_mean', 'out_sample']
        for key in possible_recons_names:
            try:
                recons = outputs[key]
                if recons is not None:
                    break  # if we found it and it's not None
            except KeyError:
                pass
        if recons is None:
            msg = ("Couldn't find reconstruction in the output dictionary. "
                   "Tried keys: {}".format(possible_recons_names))
            raise RuntimeError(msg)

        # Pick required number of images
        x = x[:n_img]
        recons = recons[:n_img]

        # Save inputs and reconstructions in a grid
        save_image_grid_reconstructions(x, recons, filename)

    def save_images(self, img_folder: str, nrows: Optional[int] = 8) -> None:
        """
        Default method to save test images for VAE-type models.

        Saves samples from the generative model, and input/reconstruction
        pairs from the test set.

        Args:
            img_folder (str): Folder to store images
            nrows (int, optional): number of rows (and columns) of images. Total
                number of sub-images in each grid will be nrows**2.
        """

        step = self.model.global_step

        # Save model samples
        fname = os.path.join(img_folder, 'sample_' + str(step) + '.png')
        self.generate_and_save_samples(fname, nrows)

        # Get first test batch
        (x, _) = next(iter(self.dataloaders.test))

        # Save model original/reconstructions
        fname = os.path.join(img_folder, 'reconstruction_' + str(step) + '.png')

        self.generate_and_save_reconstructions(x, fname, nrows)



import os

import numpy as np
import torch
from torch import nn

from boilr.nn.utils import get_module_device
from boilr.utils import checkpoints_in_folder


class BaseModel(nn.Module):
    """Base model class."""

    def __init__(self):
        super().__init__()
        self._global_step = 0

    def increment_global_step(self):
        """Increments global step by 1."""
        self._global_step += 1

    def get_device(self):
        """Returns model device."""
        return get_module_device(self)

    def checkpoint(self, ckpt_folder, max_ckpt=None):
        """Saves model checkpoint and optionally deletes old ones.

        Args:
            ckpt_folder (str): Checkpoint folder
            max_ckpt (int, optional): If not None, maximum number of most
                recent checkpoints to be kept. Older ones are deleted.
        """

        # Get checkpoints before saving the new one (is torch.save synchronous?)
        filenames, _ = checkpoints_in_folder(ckpt_folder)

        # Save checkpoint
        path = os.path.join(ckpt_folder, "model_{}.pt".format(self.global_step))
        torch.save(self.state_dict(), path)

        # Return if we're supposed to keep all checkpoints
        if max_ckpt is None or max_ckpt == -1:
            return

        # Delete all old checkpoints except for the last max_ckpt-1
        if len(filenames) < max_ckpt - 1:
            return
        for i in range(len(filenames) - max_ckpt + 1):
            path = os.path.join(ckpt_folder, filenames[i])
            try:
                os.remove(path)
            except OSError:
                pass

    def load(self, ckpt_folder, device=None, step=None):
        """Loads model from checkpoint.

        Args:
            ckpt_folder (str): Checkpoint folder
            device (torch.device, optional): Device to move the model dict to.
                Default: None. If this is None, the model's state dict is not
                moved.
            step (int, optional): Global step from which to load the
                checkpoint. Default: None. If this is None, the latest
                checkpoint is loaded.
        """
        if step is None:
            filenames, numbers = checkpoints_in_folder(ckpt_folder)
            ckpt_name = filenames[np.argmax(numbers)]  # get latest checkpoint
            step = max(numbers)
        else:
            ckpt_name = "model_{}.pt".format(step)
        print("Loading model checkpoint at step {}...".format(step))
        path = os.path.join(ckpt_folder, ckpt_name)
        self.load_state_dict(torch.load(path, map_location=device))
        self._global_step = step
        print("Loaded.")

    @property
    def global_step(self):
        """Global step."""
        return self._global_step


class BaseGenerativeModel(BaseModel):
    """Base class for generative models."""

    def sample_prior(self, n_imgs, **kwargs):
        """Samples an observation from the generative model.

        Args:
            n_imgs: Number of datapoints to be generated
        """
        raise NotImplementedError



import datetime
import os
import random
import re

import numpy as np
import torch


def set_rnd_seed(seed, aggressive=False):
    torch.manual_seed(seed)
    np.random.seed(seed)
    random.seed(seed)

    # The two lines below might slow down training
    if aggressive:
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = False


def get_date_str():
    return datetime.datetime.now().strftime("%y%m%d_%H%M%S")


def linear_anneal(x, start, end, steps):
    assert x >= 0
    assert steps > 0
    assert start >= 0
    assert end >= 0
    if x > steps:
        return end
    if x < 0:
        return start
    return start + (end - start) / steps * x


def to_np(x):
    try:
        return x.detach().cpu().numpy()
    except AttributeError:
        return x


def checkpoints_in_folder(folder):
    """Finds checkpoints in a folder.

    Finds checkpoint files in the speficied folder, return a list of file names
    and a list of integers corresponding to the time steps when the checkpoints
    were saved. These lists have the same length, they have a one-to-one
    correspondence, and they are sorted by number.

    Arguments:
        folder (str)

    Returns:
        filenames (list of str): File names of checkpoints. The full path is
            obtained by joining the input argument folder with these file
            names.
        numbers (list of int)
    """

    def is_checkpoint_file(f):
        # TODO use regex properly and maybe get number directly here
        full_path = os.path.join(folder, f)
        return (os.path.isfile(full_path) and f.startswith("model_") and
                f.endswith('.pt'))

    filenames = [f for f in os.listdir(folder) if is_checkpoint_file(f)]
    regex = re.compile(r'\d+')
    numbers = list([int(regex.search(n).group(0)) for n in filenames])
    assert len(filenames) == len(numbers)
    filenames = list([filenames[i] for i in np.argsort(numbers)])
    return filenames, numbers



import os
import pickle
import timeit
import warnings

import torch
try:
    from torch.utils import tensorboard
    have_tensorboard = True
except ImportError as e:
    msg = "{}: {}".format(e.__class__.__name__, e)
    msg = "Could not import tensorboard.\n" + msg
    warnings.warn(msg)
    have_tensorboard = False
from tqdm import tqdm

from boilr.nn.utils import global_norm, grad_global_norm
from boilr.options import get_option
from boilr.utils import set_rnd_seed, get_date_str
from boilr.utils import viz
from boilr.utils.summarize import History, SummarizerCollection


class Trainer:
    """Generic tool for training models.

    All model- and experiment-specific work is defined and performed in an
    experiment object, which is provided at initialization.

    Args:
        experiment (boilr.experiments.BaseExperimentManager)
    """

    def __init__(self, experiment):
        self.experiment = experiment
        resume = experiment.args.resume != ""

        if resume:
            # Folder string = run name = resume argument
            folder_str = experiment.args.resume
            print("Resume from '{}'".format(folder_str))
            msg = "When resuming training, the optimizer's state is not restored"
            warnings.warn(msg)
            msg = ("When resuming training, the relative time on tensorboard "
                   "has a gap due to wall clock timestamps")
            warnings.warn(msg)

            # Get all folder names to resume saving results
            tboard_folder = self._setup_paths(folder_str)
            self.tb_writer = None
            if have_tensorboard:  # maybe we didn't have tboard originally
                os.makedirs(tboard_folder, exist_ok=True)
                self.tb_writer = tensorboard.SummaryWriter(tboard_folder)

            # Forget about all arguments, load all of them from saved config
            # (the 'resume' argument is overwritten in the process)
            config_path = os.path.join(self.checkpoint_folder, 'config.pkl')
            experiment.load_args_from_pickle(config_path)

            # Load training and test history from log.pkl
            with open(self.log_path, 'rb') as file:
                history = pickle.load(file)
                self.train_history = History(history['train'])
                self.test_history = History(history['test'])

        args = experiment.args

        assert args.checkpoint_every % args.test_log_every == 0

        # Pick device (cpu/cuda)
        use_cuda = not args.no_cuda and torch.cuda.is_available()
        self.device = torch.device("cuda" if use_cuda else "cpu")

        if not resume:

            # To save training and test metrics
            self.train_history = History()
            self.test_history = History()

            # Random seed
            set_rnd_seed(args.seed)

            # Get starting time and date for logging
            date_str = get_date_str()

            # Print info
            print('Device: {}, start time: {}'.format(self.device, date_str))

            # Create folders for logs and results, save config
            folder_str = date_str
            run_descr = experiment.run_description
            if len(run_descr) > 0:
                folder_str += '_' + experiment.run_description
            tboard_folder = self._setup_paths(folder_str)
            self.tb_writer = None
            if not args.dry_run:
                os.makedirs(self.img_folder)
                os.makedirs(self.checkpoint_folder)
                if have_tensorboard:
                    os.makedirs(tboard_folder)
                    self.tb_writer = tensorboard.SummaryWriter(tboard_folder)
                config_path = os.path.join(self.checkpoint_folder, 'config.pkl')
                with open(config_path, 'wb') as fd:
                    pickle.dump(args, fd)

        viz.img_folder = self.img_folder
        experiment.setup(
            device=self.device,
            checkpoint_folder=self.checkpoint_folder if resume else None)

        # Check everything is initialized properly
        self._check_experiment(experiment)

    def _setup_paths(self, folder_str):
        result_folder = os.path.join('output', folder_str, 'results')
        self.img_folder = os.path.join(result_folder, 'imgs')
        self.checkpoint_folder = os.path.join('output', folder_str,
                                              'checkpoints')
        self.log_path = os.path.join(result_folder, 'log.pkl')
        tboard_folder = os.path.join('output', folder_str, 'tensorboard_logs')
        return tboard_folder

    def run(self):
        """Runs the trainer."""

        # This is needed when loading to resume training
        first_step = True

        # Setup
        e = self.experiment
        train_loader = e.dataloaders.train
        progress = None
        show_progress = get_option('show_progress_bar')
        train_summarizers = SummarizerCollection(
            mode='moving_average',
            ma_length=get_option('train_summarizer_ma_length'))
        # Additional summarizers are considered independent of the train/test
        # regime, they are not printed, they are saved to tensorboard only once
        # (during training and not testing), and for now they are not saved to
        # the history.
        additional_summarizers = SummarizerCollection(
            mode='moving_average',
            ma_length=get_option('train_summarizer_ma_length'))

        # Training mode
        e.model.train()

        # Main loop
        for epoch in range(1, e.args.max_epochs + 1):
            for batch_idx, (x, y) in enumerate(train_loader):

                step = e.model.global_step

                if step >= e.args.max_steps:
                    break

                if step % e.args.test_log_every == 0:

                    # Test model (unless we just resumed training)
                    if not first_step or step == 0:
                        with torch.no_grad():
                            self._test(epoch)

                    # Save model checkpoint (unless we just started/resuming training)
                    if not first_step and step % e.args.checkpoint_every == 0:
                        print("* saving model checkpoint at "
                              "step {}".format(step))
                        e.model.checkpoint(self.checkpoint_folder,
                                           e.args.keep_checkpoint_max)

                    # (Re)start progress bar
                    if show_progress:
                        progress = tqdm(total=e.args.test_log_every,
                                        desc='train')

                    # Restart timer to measure training speed
                    timer_start = timeit.default_timer()
                    steps_start = e.model.global_step

                    # This timer stuff won't make sense if 'test every' is not
                    # a multiple of 'train every'. Which is now true, but let's
                    # be safe in case things change
                    assert e.args.test_log_every % e.args.test_log_every == 0

                # Reset gradients
                e.optimizer.zero_grad()

                # Forward pass: get loss and other info
                outputs = e.forward_pass(x, y)

                # Compute gradients (backward pass)
                outputs['loss'].backward()

                e.post_backward_callback()

                if e.args.max_grad_norm is not None:
                    torch.nn.utils.clip_grad_norm_(
                        e.model.parameters(), max_norm=e.args.max_grad_norm)

                # Add batch metrics to summarizers
                metrics_dict = e.get_metrics_dict(outputs)
                train_summarizers.add(metrics_dict)

                # Compute gradient stats and add to summarizers
                # - grad norm of each parameter
                # - grad norm of given group (or default groups)
                # - total grad norm
                # TODO groups
                grad_stats = {
                    'grad_norm_total/grad_norm_total':
                        grad_global_norm(e.model.parameters())
                }
                for n, p in e.model.named_parameters():
                    k = 'grad_norm_per_parameter/' + n
                    grad_stats[k] = grad_global_norm(p)
                additional_summarizers.add(grad_stats)

                # Compute L2 norm of parameters and add it to summarizers
                additional_summarizers.add(
                    {'L2_norm': global_norm(e.model.parameters())})

                # Update progress bar
                if show_progress:
                    progress.update()

                # Close progress bar if test occurs at next loop iteration
                if (step + 1) % e.args.test_log_every == 0:
                    # If show_progress is False, progress is None
                    if progress is not None:
                        progress.close()

                if (step + 1) % e.args.train_log_every == 0:
                    # step+1 because we already did a forward/backward step

                    # Get moving average of metrics and reset summarizers
                    train_summaries = train_summarizers.get_all(reset=True)
                    additional_summaries = additional_summarizers.get_all(
                        reset=True)

                    # Get training speed and add it to summaries
                    elapsed = timeit.default_timer() - timer_start
                    iterations = e.model.global_step - steps_start
                    steps_per_sec = iterations / elapsed
                    ex_per_sec = steps_per_sec * e.args.batch_size
                    additional_summaries['speed/steps_per_sec'] = steps_per_sec
                    additional_summaries['speed/examples_per_sec'] = ex_per_sec
                    timer_start = timeit.default_timer()
                    steps_start = e.model.global_step

                    # Print summaries
                    print(e.train_log_str(train_summaries, step + 1, epoch))

                    # Add train summaries (smoothed) to history and dump it to
                    # file and to tensorboard if available
                    self.train_history.add(train_summaries, step + 1)
                    if not e.args.dry_run:
                        with open(self.log_path, 'wb') as fd:
                            pickle.dump(self._history_dict(), fd)
                        if self.tb_writer is not None:
                            for k, v in train_summaries.items():
                                self.tb_writer.add_scalar(
                                    'train_' + k, v, step + 1)
                            for k, v in additional_summaries.items():
                                self.tb_writer.add_scalar(k, v, step + 1)

                # Optimization step
                e.optimizer.step()

                # Increment model's global step variable
                e.model.increment_global_step()

                first_step = False

            if step >= e.args.max_steps:
                break

        if progress is not None:  # if show_progress is False, progress is None
            progress.close()
        if step < e.args.max_steps:
            print("Reached epochs limit ({} epochs)".format(e.args.max_epochs))
        else:
            print("Reached steps limit ({} steps)".format(e.args.max_steps))

    def _test(self, epoch):
        e = self.experiment
        step = e.model.global_step

        # Evaluation mode
        e.model.eval()

        # Get test results
        summaries = e.test_procedure()

        # Save images
        if step % e.args.test_imgs_every == 0:
            if not e.args.dry_run:
                e.save_images(self.img_folder)

        # Print log string with test results (experiment-specific)
        print(e.test_log_str(summaries, step, epoch))

        # Save summaries to history
        self.test_history.add(summaries, step)

        if not e.args.dry_run:

            # Save summaries to tensorboard
            if self.tb_writer is not None:
                for k, v in summaries.items():
                    self.tb_writer.add_scalar('validation_' + k, v, step)

            # Save history to file
            with open(self.log_path, 'wb') as fd:
                pickle.dump(self._history_dict(), fd)

        # Training mode
        e.model.train()

    def _history_dict(self):
        return {
            'train': self.train_history.get_dict(),
            'test': self.test_history.get_dict(),
        }

    @staticmethod
    def _check_experiment(e):
        attributes = [
            e.device,
            e.dataloaders,
            e.model,
            e.args,
            e.run_description,
            e.optimizer,
        ]
        assert not (None in attributes)



from collections import deque

import numpy as np


class History:
    """Collection of summarizers, useful as history of training metrics.

    It wraps SummarizerCollection to register new history values with a
    timestamp.

    Args:
        init_history (SummarizerCollection, optional)
    """

    def __init__(self, init_history=None):
        if init_history is None:
            self._summarizers = SummarizerCollection(mode='store_all')
        else:
            msg = ("provided initialization argument has to be "
                   "an instance of SummarizerCollection")
            assert isinstance(init_history, SummarizerCollection), msg
            self._summarizers = init_history

    def add(self, summaries, timestamp):
        """Adds new summaries with a timestamp.

        Args:
            summaries (dict): Dictionary of summaries. E.g. key is the name
                of the quantity, value is its average over the last epoch.
            timestamp: Timestamp, typically the global step.
        """
        timestamped_dict = {k: (timestamp, v) for k, v in summaries.items()}
        self._summarizers.add(timestamped_dict)

    def get_dict(self):
        """Returns this history as dictionary."""
        return self._summarizers

    def __str__(self):
        return str(self._summarizers)

    def __repr__(self):
        return repr(self._summarizers)


class SummarizerCollection:
    """Collection of summarizers.

    Args:
        mode (str): Type of summarizers: 'total_average', 'sum',
            'moving_average', 'store_all'.
        ma_length (int, optional): If the mode is 'moving_average', this is
            the size of the moving average window, and it is not optional.
    """

    def __init__(self, mode, ma_length=None):
        legal_modes = [
            'total_average',
            'sum',
            'moving_average',
            'store_all',
        ]
        if mode not in legal_modes:
            raise RuntimeError("unrecognized mode '{}'".format(mode))
        assert (ma_length is not None) == (mode == 'moving_average')
        self.ma_length = ma_length
        self.mode = mode
        self.summarizers = {}  # dict of summarizers

    def _new_summarizer(self):
        """Creates a new summarizer for the collection."""
        if self.mode == 'store_all':
            return HistorySummarizer()
        if self.mode == 'total_average':
            return TotalAverageSummarizer()
        if self.mode == 'sum':
            return SumSummarizer()
        if self.mode == 'moving_average':
            return MASummarizer(self.ma_length)

    def add(self, data):
        """Update the summarizers with new values.

        Args:
            data (dict)
        """

        for name, value in data.items():
            if name not in self.summarizers:
                self.summarizers[name] = self._new_summarizer()
            self.summarizers[name].add(value)

    def get(self, name):
        """Returns the summary of the quantity with the given name."""
        return self.summarizers[name].get()

    def get_all(self, reset=False):
        """Returns a dictionary of all summaries.

        The keys are the names of the quantities, and the values are the
        corresponding summaries. Optionally, the summarizer can be reset.
        Note: the dictionary is a shallow copy of the internal dict.

        Args:
            reset (bool, optional): Whether the summarizers should be reset.

        Returns:
            dictionary (dict)
        """
        out = {n: self[n] for n in self}
        if reset:
            self.reset()
        return out

    def __getitem__(self, name):
        return self.get(name)

    def __contains__(self, name):
        return name in self.summarizers

    def __iter__(self):
        for name in self.summarizers:
            yield name

    def __str__(self):
        s = type(self).__name__ + "(\n"
        s += "  mode: {}\n".format(self.mode)
        if self.ma_length is not None:
            s += "  ma_length: {}\n".format(self.ma_length)
        s += "  summarizers:\n"
        for name in self:
            s += "    {}: {}\n".format(name, self.summarizers[name])
        return s + ")"

    def reset(self):
        """Re-initializes the object.

        This should be called after getting the summary of all quantities,
        and before starting to collect the new values for the next summary.
        """
        self.__init__(self.mode, self.ma_length)


class Summarizer:
    """Base class for Summarizers.

    A Summarizer provides a simple way to get a summary of a sequence of
    values. The specific summary depends on the implementation of the
    Summarizer.

    A Summarizer is updated by adding new values to it. It keeps a state that
    represents the summary of values added so far, so the method get() simply
    returns the internal state.

    For now, a summary can be the full history, the overall average, a moving
    average with a limited window, or the sum of all values.
    """

    def __init__(self):
        self.state = None

    def add(self, value):
        """Adds a value to the summarizer.

        Args:
            value: Value to be added
        """
        pass

    def get(self):
        """Returns summarizer's state."""
        return self.state


class HistorySummarizer(Summarizer):
    """History summarizer. Stores the history of all values."""

    def __init__(self):
        super().__init__()
        self.state = []

    def add(self, value):
        """See base class."""
        self.state.append(value)

    def __str__(self):
        if len(self.state) < 4:
            return ", ".join(str(v) for v in self.state)
        return "{}, ..., {}".format(self.state[0], self.state[-1])


class TotalAverageSummarizer(Summarizer):
    """Summarizer that stores the total average of all values ever added."""

    def __init__(self):
        super().__init__()
        self.n_items = 0
        self.state = 0.0

    def add(self, value):
        """See base class."""
        self.n_items += 1
        self.state += (value - self.state) / self.n_items

    def __str__(self):
        return "{} ({} items)".format(self.state, self.n_items)


class SumSummarizer(Summarizer):
    """Summarizer that stores the total sum of all values ever added."""

    def __init__(self):
        super().__init__()
        self.n_items = 0
        self.state = 0.0

    def add(self, value):
        """See base class."""
        self.n_items += 1
        self.state += value

    def __str__(self):
        return "{} ({} items)".format(self.state, self.n_items)


class MASummarizer(Summarizer):
    """Summarizer that stores the moving average with a fixed window.

    Args:
        ma_length (int): Size of the moving average window
    """

    def __init__(self, ma_length):
        super().__init__()
        self.ma_length = ma_length
        self.list = deque(maxlen=ma_length)
        self.state = 0.0

    def add(self, value):
        """See base class."""
        n = len(self.list)
        if n == self.ma_length:
            self.state += (value - self.list[0]) / n
        else:  # n < ma_length
            # always keeps average of numbers in memory, even
            # when there's fewer than ma_length
            self.state += (value - self.state) / (n + 1)
        self.list.append(value)

    def __str__(self):
        return "{}".format(self.state)


if __name__ == '__main__':
    summ = MASummarizer(4)
    for i in range(10):
        summ.add(3)
        assert summ.get() == 3
        assert len(summ.list) == min(i + 1, summ.ma_length)

    summ = TotalAverageSummarizer()
    for i in range(10):
        summ.add(3)
        assert summ.get() == 3
        assert summ.n_items == i + 1

    summ = MASummarizer(20)
    for i in range(50):
        summ.add(np.random.rand())
        assert abs(summ.get() - np.mean(summ.list)) < 1e-8, \
            "{} != {}".format(summ.get(), np.mean(summ.list))
        assert len(summ.list) == min(i + 1, summ.ma_length)



"""Package-wide options."""

from copy import deepcopy

# Define options and set default values
_options = {
    'model_print_depth': 3,
    'train_summarizer_ma_length': 1000,
    'show_progress_bar': True,
}


def set_options(**kwargs):
    """Sets specified package-wide options.

    Args:
        **kwargs: keyword arguments defining options that have to be updated.
            Options that are not specified are left unchanged.
    """

    # Check arguments
    invalid = []
    for k in kwargs:
        if k not in _options:
            invalid.append(k)
    if len(invalid) > 0:
        msg = "Invalid option names: "
        invalid = [repr(k) for k in invalid]
        msg = msg + ", ".join(invalid)
        raise ValueError(msg)

    for k, v in kwargs.items():
        if not _check_option(k, v):
            msg = "Invalid value {} for option {}".format(repr(v), repr(k))
            raise ValueError(msg)

    # Update options
    _options.update(**kwargs)


def get_option(name):
    """Returns the current value of the specified option.

    Args:
        name (str): Option name

    Returns:
        value: Option value
    """
    return _get_options(name)


def get_options():
    """Returns the current value of all options.

    Returns:
        options (dict): all packagewide options
    """
    return _get_options()


def _get_options(name=None):
    """Returns the current value of either one or all options.

    Args:
        name (str, optional): Name of the required option

    Returns:
        If name is specified, the value of the option with the specified name.
        Otherwise, dictionary containing all options.
    """
    if name is None:
        return deepcopy(_options)
    return deepcopy(_options[name])


def _check_option(name, value):
    """Check that the value is valid for the option with the given name.

    Args:
        name (str): Option name
        value: Option value

    Returns:
        True if ``value`` is a valid value for option ``name``, False
        otherwise.

    Raises:
        ValueError: Unknown option name.
    """
    if name == 'model_print_depth':
        if value is None:
            return True
        return isinstance(value, int) and value > 0
    elif name == 'train_summarizer_ma_length':
        return isinstance(value, int) and value > 0
    elif name == 'show_progress_bar':
        return isinstance(value, bool)
    else:
        raise ValueError("Unknown option name {}".format(repr(name)))



"""Useful basic layers and ops for neural nets."""

from torch import nn
from torch.nn import functional as F


class Interpolate(nn.Module):
    """Wrapper for torch.nn.functional.interpolate."""

    def __init__(self,
                 size=None,
                 scale=None,
                 mode='bilinear',
                 align_corners=False):
        super().__init__()
        assert (size is None) == (scale is not None)
        self.size = size
        self.scale = scale
        self.mode = mode
        self.align_corners = align_corners

    def forward(self, x):
        out = F.interpolate(x,
                            size=self.size,
                            scale_factor=self.scale,
                            mode=self.mode,
                            align_corners=self.align_corners)
        return out


class CropImage(nn.Module):
    """Crops image to given size.

    Args:
        size
    """

    def __init__(self, size):
        super().__init__()
        self.size = size

    def forward(self, x):
        return crop_img_tensor(x, self.size)


class Reshape(nn.Module):

    def __init__(self, *args, implicit_batch=True, allow_copy=False):
        super().__init__()
        self.shape = args
        self.implicit_batch = implicit_batch
        self.allow_copy = allow_copy

    def forward(self, x):
        shp = self.shape
        if self.implicit_batch:
            shp = (x.size(0), *shp)
        if self.allow_copy:
            return x.reshape(shp)
        return x.view(shp)


class PrintShape(nn.Module):

    def __init__(self):
        super().__init__()
        self.first_pass = True

    def forward(self, x):
        if self.first_pass:
            print(" > shape:", x.shape)
            self.first_pass = False
        return x


class Identity(nn.Module):

    def forward(self, x):
        return x


def pad_img_tensor(x, size):
    """Pads a tensor.

    Pads a tensor of shape (batch, channels, h, w) to new height and width
    given by a tuple.

    Args:
        x (torch.Tensor): Input image
        size (iterable): Desired size (height, width)

    Returns:
        The padded tensor
    """

    return _pad_crop_img(x, size, 'pad')


def crop_img_tensor(x, size):
    """Crops a tensor.

    Crops a tensor of shape (batch, channels, h, w) to new height and width
    given by a tuple.

    Args:
        x (torch.Tensor): Input image
        size (iterable): Desired size (height, width)

    Returns:
        The cropped tensor
    """
    return _pad_crop_img(x, size, 'crop')


def _pad_crop_img(x, size, mode):
    """ Pads or crops a tensor.

    Pads or crops a tensor of shape (batch, channels, h, w) to new height
    and width given by a tuple.

    Args:
        x (torch.Tensor): Input image
        size (iterable): Desired size (height, width)
        mode (str): Mode, either 'pad' or 'crop'

    Returns:
        The padded or cropped tensor
    """

    assert x.dim() == 4 and len(size) == 2
    size = tuple(size)
    x_size = x.size()[2:4]
    if mode == 'pad':
        cond = x_size[0] > size[0] or x_size[1] > size[1]
    elif mode == 'crop':
        cond = x_size[0] < size[0] or x_size[1] < size[1]
    else:
        raise ValueError("invalid mode '{}'".format(mode))
    if cond:
        raise ValueError('trying to {} from size {} to size {}'.format(
            mode, x_size, size))
    dr, dc = (abs(x_size[0] - size[0]), abs(x_size[1] - size[1]))
    dr1, dr2 = dr // 2, dr - (dr // 2)
    dc1, dc2 = dc // 2, dc - (dc // 2)
    if mode == 'pad':
        return nn.functional.pad(x, [dc1, dc2, dr1, dr2, 0, 0, 0, 0])
    elif mode == 'crop':
        return x[:, :, dr1:x_size[0] - dr2, dc1:x_size[1] - dc2]


def free_bits_kl(kl, free_bits, batch_average=False, eps=1e-6):
    """Computes free-bits version of KL divergence.

    Takes in the KL with shape (batch size, layers), returns the KL with
    free bits (for optimization) with shape (layers,), which is the average
    free-bits KL per layer in the current batch.

    If batch_average is False (default), the free bits are per layer and
    per batch element. Otherwise, the free bits are still per layer, but
    are assigned on average to the whole batch. In both cases, the batch
    average is returned, so it's simply a matter of doing mean(clamp(KL))
    or clamp(mean(KL)).

    Args:
        kl (torch.Tensor)
        free_bits (float)
        batch_average (bool, optional))
        eps (float, optional)

    Returns:
        The KL with free bits
    """

    assert kl.dim() == 2
    if free_bits < eps:
        return kl.mean(0)
    if batch_average:
        return kl.mean(0).clamp(min=free_bits)
    return kl.clamp(min=free_bits).mean(0)
